unit ComIO;
{ COM port unit v1.01 beta         Copyright (c) 1997, Marius Gedminas }
{                                                                      }
{ This unit implements communication with a serial port. Any number of }
{ TComPort object may be active at a time, non-standard port addresses }
{ may be used, but only IRQ3 and IRQ4 are now supported (it would be   }
{ easy to extend this). Two COM ports with the same IRQ can be used at }
{ the same time (I had no chance to test this, but it should work -    }
{ theoretically).                                                      }
{                                                                      }
{ This unit is free non-commercial use. You may change the code below  }
{ for your own use, but don't give away modified versions of it. If    }
{ you think you made an improvement, please contact me at              }
{ mgedmin@pub.osf.lt                                                   }
{                                                                      }
{ Warning: this unit wasn't tested very thoroughly. I think it should  }
{ be used only as an example, but if you wish you can risk and use it  }
{ in your programs.                                                    }
{                                                                      }
{ The code is poorly commented, perhaps it will change in the future.  }
{                                                                      }
{ A short disclaimer: use this at your own risk. I won't be liable for }
{ ANYTHING concerning this unit, and I don't give ANY warranties.      }

{ Some notes:                                                          }
{                                                                      }
{ Three modes of operation are supported:                              }
{   comInt14 - standard BIOS INT 14 interface - DON'T USE IT - it      }
{              doesn't work yet - and I think I'll remove this later - }
{              this is the worst method to access serial ports         }
{   comPoll  - polling - I don't think it is very reliable unless you  }
{              check for arriving characters often                     }
{   comInt   - interrupt-driven - this should me used most of time     }
{                                                                      }
{ Things that doesn't work:                                            }
{   comInt14 mode                                                      }
{   Software flow control                                              }
{   IRQs other than 3 or 4                                             }
{   Speed beyond 9600 bps doesn't work (at least on my PC: 16450 UART  }
{     and 2400 bps modem)                                              }
{                                                                      }
{ Example of usage:                                                    }
{   var                                                                }
{     P: TComPort;                                                     }
{   ...                                                                }
{   P.Init(2, comInt, 9600, 8, ParityNone, 1);                         }
{                               { COM2 at 9600 bps, line settings: 8n1 }
{   if P.Pending then         { A character is available in the buffer }
{     P.Receive(B);                                           { Get it }
{   P.Send(B);                                      { And send it back }
{   { You could check if the port is ready to receive a character:     }
{   { repeat ... until P.Ready;                                        }
{   { But this isn't necessary - P.Send will do it for you             }
{   if P.Error <> 0 then                          { Some error occured }
{     ...                                                              }
{   P.Done;                                        { Finished at last! }

{ Changes for 1.01 beta:                                               }
{   AbortIO function                                                   }
{   GetTimer is now in the interface section                           }
{   Bps is now Longint (115200 doesn't fit in a word)                  }
{   BUG: Send and Receive returned True even if Error wasn't errOk     }
{   BUG: Send: Time was Word, not Longint - caused some timeouts       }

interface

  uses
    Dos;

  const
    MaxBufSize = 65528;
    StdBufSize: Word = 4096;
    StdTimeout: Word = 2*18; { 2 seconds }

  type
    WordRec = record
      Lo, Hi: Byte;
    end;

    TPortNo = 1..4;
    TComMode = (comInt14, comPoll, comInt);
    TParity = (ParityNone, ParityOdd, ParityEven);
    PComBuffer = ^TComBuffer;
    TComBuffer = array[0..MaxBufSize-1] of Byte;

    TAbortIO = function: Boolean;

    PComPort = ^TComPort;
    TComPort = object
      Next: PComPort;           { read only }
      { Hardware }
      PortNo: Byte;             { read only }
      PortBase: Word;           { read only }
      IrqNo: Byte;              { read only }
      { Connection }
      Bps: Longint;             { read only }
      Data: Byte;               { read only }
      Stop: Byte;               { read only }
      Parity: TParity;          { read only }
      OldIER, OldMCR, OldLCR,   { read only }
      OldMSR, OldDLL, OldDLH: Byte;
      { Mode }
      Mode: TComMode;           { must switch int/other modes with InitInt/DoneInt }
      Buffer: PComBuffer;       { read only }
      BufSize: Word;            { read only }
      Head, Tail: Word;         { read only }
      { Flow control }
      HardwareFlow, SoftwareFlow: Boolean; { SoftwareFlow not supported yet }
      Timeout: Word;
      { Error handling }
      Error: Integer;
      AbortIO: TAbortIO;

      procedure Init(APortNo: TPortNo; AMode: TComMode;
                  ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
      procedure InitCustom(APortBase: Word; AIrqNo: Byte; AMode: TComMode;
                  ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
      procedure Setup(ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
      procedure InitInt(ABufSize: Word);
      procedure DoneInt;
      procedure Done;

      function LineStatus: Byte;
      function ModemStatus: Byte;
      function Ready: Boolean;
      function Send(B: Byte): Boolean;
      function Pending: Boolean;
      function Receive(var B: Byte): Boolean;
    end;

  const
    { Error codes }
    errOk               = 0;
    errBadSettings      = 1;
    errCantSend         = 2;
    errTimeout          = 3;
    errOverflow         = 4;
    { Line status }
    lsDataReady         = $01; { Data ready. }
    lsOverrunError      = $02; { Overrun error - Data register overwritten. }
    lsParityError       = $04; { Parity error - bad transmission. }
    lsFramingError      = $08; { Framing error - No stop bit was found. }
    lsBreakDetect       = $10; { Break detect - End to transmission requested. }
    lsTrHoldEmpty       = $20; { Transmitter holding register is empty. }
    lsTrShiftEmpty      = $40; { Transmitter shift register is empty. }
    lsTimeout           = $80; { Time out - off line. }
    { Modem status }
    msDeltaCTS          = $01; { delta Clear To Send. }
    msDeltaDSR          = $02; { delta Data Set Ready. }
    msDeltaRI           = $04; { delta Ring Indicator. }
    msDeltaDCD          = $08; { delta Data Carrier Detect. }
    msCTS               = $10; { Clear To Send. }
    msDSR               = $20; { Data Set Ready. }
    msRI                = $40; { Ring Indicator. }
    msDCD               = $80; { Data Carrier Detect. }

  function GetTimer: Longint;

implementation

const
  StdIrqNo: array[TPortNo] of Byte = (4, 3, 4, 3);
var
  StdPortBase: array[TPortNo] of Word absolute $0040:$0000;
  { = ($3F8, $2F8, $2E8, $2F8); }

const
  {
      The 8250 UART has 10 registers accessible through 7 port addresses.
      Here are their addresses relative to COM1BASE and COM2BASE. Note
      that the baud rate registers, (DLL) and (DLH) are active only when
      the Divisor-Latch Access-Bit (DLAB) is on. The (DLAB) is bit 7 of
      the (LCR).

          o TXR Output data to the serial port.
          o RXR Input data from the serial port.
          o LCR Initialize the serial port.
          o IER Controls interrupt generation.
          o IIR Identifies interrupts.
          o MCR Send contorl signals to the modem.
          o LSR Monitor the status of the serial port.
          o MSR Receive status of the modem.
          o DLL Low byte of baud rate divisor.
          o DHH High byte of baud rate divisor.
  }
    TXR        =     0;       {  Transmit register (WRITE) }
    RXR        =     0;       {  Receive register  (READ)  }
    IER        =     1;       {  Interrupt Enable          }
    IIR        =     2;       {  Interrupt ID              }
    LCR        =     3;       {  Line control              }
    MCR        =     4;       {  Modem control             }
    LSR        =     5;       {  Line Status               }
    MSR        =     6;       {  Modem Status              }
    DLL        =     0;       {  Divisor Latch Low         }
    DLH        =     1;       {  Divisor latch High        }


  {-------------------------------------------------------------------*
    Bit values held in the Line Control Register (LCR).
          bit           meaning
          ---           -------
          0-1           00=5 bits, 01=6 bits, 10=7 bits, 11=8 bits.
          2             Stop bits.
          3             0=parity off, 1=parity on.
          4             0=parity odd, 1=parity even.
          5             Sticky parity.
          6             Set break.
          7             Toggle port addresses.
   *-------------------------------------------------------------------}
    NO_PARITY   =    $00;
    EVEN_PARITY =    $18;
    ODD_PARITY  =    $08;



  {-------------------------------------------------------------------*
    Bit values held in the Line Status Register (LSR).
          bit           meaning
          ---           -------
          0             Data ready.
          1             Overrun error - Data register overwritten.
          2             Parity error - bad transmission.
          3             Framing error - No stop bit was found.
          4             Break detect - End to transmission requested.
          5             Transmitter holding register is empty.
          6             Transmitter shift register is empty.
          7               Time out - off line.
   *-------------------------------------------------------------------}
    RCVRDY      =    $01;
    OVRERR      =    $02;
    PRTYERR     =    $04;
    FRMERR      =    $08;
    BRKERR      =    $10;
    XMTRDY      =    $20;
    XMTRSR      =    $40;
    TIMEOUT     =    $80;
    ZER         =    $00;
  {-------------------------------------------------------------------*
    Bit values held in the Modem Output Control Register (MCR).
          bit           meaning
          ---           -------
          0             Data Terminal Ready. Computer ready to go.
          1             Request To Send. Computer wants to send data.
          2             auxillary output #1.
          3             auxillary output #2.(Note: This bit must be
                          set to allow the communications card to send
                          interrupts to the system)
          4             UART ouput looped back as input.
          5-7           not used.
   *------------------------------------------------------------------}
    DTR         =    $01;
    RTS         =    $02;
    MC_INT      =    $08;


  {------------------------------------------------------------------*
    Bit values held in the Modem Input Status Register (MSR).
          bit           meaning
          ---           -------
          0             delta Clear To Send.
          1             delta Data Set Ready.
          2             delta Ring Indicator.
          3             delta Data Carrier Detect.
          4             Clear To Send.
          5             Data Set Ready.
          6             Ring Indicator.
          7             Data Carrier Detect.
   *------------------------------------------------------------------}
    CTS         =    $10;
    DSR         =    $20;


  {------------------------------------------------------------------*
    Bit values held in the Interrupt Enable Register (IER).
          bit           meaning
          ---           -------
          0             Interrupt when data received.
          1             Interrupt when transmitter holding reg. empty.
          2             Interrupt when data reception error.
          3             Interrupt when change in modem status register.
          4-7           Not used.
   *------------------------------------------------------------------}
    RX_INT      =    $01;


  {------------------------------------------------------------------*
    Bit values held in the Interrupt Identification Register (IIR).
          bit           meaning
          ---           -------
          0             Interrupt pending
          1-2             Interrupt ID code
                          00=Change in modem status register,
                          01=Transmitter holding register empty,
                          10=Data received,
                          11=reception error, or break encountered.
          3-7           Not used.
   *------------------------------------------------------------------}
    RX_ID       =    $04;
    RX_MASK     =    $07;


  {
      These are the port addresses of the 8259 Programmable Interrupt
      Controller (PIC).
  }
    IMR         =    $21;   { Interrupt Mask Register port }
    ICR         =    $20;   { Interrupt Control Port       }


  {
      An end of interrupt needs to be sent to the Control Port of
      the 8259 when a hardware interrupt ends.
  }
    EOI         =    $20;   { End Of Interrupt }


  {
      The (IMR) tells the (PIC) to service an interrupt only if it
      is not masked (FALSE).
  }
    IRQ3        =    $F7;  { COM2 & COM4 }
    IRQ4        =    $EF;  { COM1 & COM3 }


const
  Irq3Ports: PComPort = nil;
  Irq4Ports: PComPort = nil;
  OldIsr3: Pointer = nil;
  OldIsr4: Pointer = nil;
  OldIsr3Enable: Boolean = False;
  OldIsr4Enable: Boolean = False;

procedure EnableIrq(No: Byte; Enable: Boolean);
begin
  if Enable then
    Port[$21] := Port[$21] and not ($01 shl No)
  else
    Port[$21] := Port[$21] or ($01 shl No);
end;

function IrqEnabled(No: Byte): Boolean;
begin
  IrqEnabled := Port[$21] and ($01 shl No) = 0;
end;

procedure ChainInt(OldInt: Pointer);
inline($5B/                { pop  bx           }
       $58/                { pop  ax           }
       $89/$EC/            { mov  sp, bp       }
       $87/$46/$10/        { xchg [bp+10h], ax }
       $87/$5E/$0E/        { xchg [bp+0Eh], bx }
       $5D/                { pop  bp           }
       $07/                { pop  es           }
       $1F/                { pop  ds           }
       $5F/                { pop  di           }
       $5E/                { pop  si           }
       $5A/                { pop  dx           }
       $59/                { pop  cx           }
       $CB);               { retf              }

procedure AddToList(var List: PComPort; APort: PComPort);
begin
  APort^.Next := List;
  List := APort;
end;

procedure RemoveFromList(var List: PComPort; APort: PComPort);
var
  P: ^PComPort;
begin
  P := @List;
  while (P^ <> APort) and (P^ <> nil) do P := @P^^.Next;
  P^ := APort^.Next;
  APort^.Next := nil;
end;

procedure Isr3; interrupt;
var
  P: PComPort;
  B: Byte;
begin
  P := Irq3Ports;
  while P <> nil do
    with P^ do
      begin
        if Port[PortBase + IIR] and RX_MASK = RX_ID then
          begin
            B := Port[PortBase + RXR];
            Buffer^[Tail] := B;
            Inc(Tail); if Tail = BufSize then Tail := 0;
            if Tail = Head then
              begin
                Error := errOverflow;
                if Tail = 0 then Tail := BufSize; Dec(Tail);
              end;
          end;
        P := Next;
      end;
  if OldIsr3Enable then
    ChainInt(OldIsr3);
  Port[$20] := $20;
end;

procedure Isr4; interrupt;
var
  P: PComPort;
  B: Byte;
begin
  P := Irq4Ports;
  while P <> nil do
    with P^ do
      begin
        if Port[PortBase + LSR] and RCVRDY <> 0 then
          begin
            B := Port[PortBase + RXR];
            Inc(Tail); if Tail = BufSize then Tail := 0;
            if Tail = Head then
              begin
                Error := errOverflow;
                if Tail = 0 then Tail := BufSize; Dec(Tail);
              end
            else
              Buffer^[Tail] := B;
          end;
        P := Next;
      end;
  if OldIsr4Enable then
    ChainInt(OldIsr4);
end;

function GetTimer: Longint; assembler;
asm
  mov  es, Seg0040
  cli
  mov  ax, es:[6Ch]
  mov  dx, es:[6Ch+2]
  sti
end;

procedure TComPort.Init(APortNo: TPortNo; AMode: TComMode;
            ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
begin
  @AbortIO := nil;
  InitCustom(StdPortBase[APortNo], StdIrqNo[APortNo], AMode,
    ABps, AData, AParity, AStop);
  PortNo := APortNo;
end;

procedure TComPort.InitCustom(APortBase: Word; AIrqNo: Byte; AMode: TComMode;
            ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
begin
  Error := errOk;
  PortNo := 0;
  PortBase := APortBase;
  IrqNo := AIrqNo;
  if AMode = comInt then Mode := comPoll else Mode := AMode;
  HardwareFlow := True;
  SoftwareFlow := False;
  Timeout := StdTimeout;
  { Get current connection settings }
  asm cli end;
  OldIER := Port[PortBase + IER];
  OldMCR := Port[PortBase + MCR];
  OldLCR := Port[PortBase + LCR];
  OldMSR := Port[PortBase + MSR];
  Port[PortBase + LCR] := $80;                     { Set DLAB }
  OldDLL := Port[PortBase + DLL];
  OldDLH := Port[PortBase + DLH];
  Port[PortBase + LCR] := OldLCR;                  { Set DLAB }
  asm sti end;
  Setup(ABps, AData, AParity, AStop);
  if AMode = comInt then
    begin
      InitInt(StdBufSize);
    end;
end;

procedure TComPort.Setup(ABps: Longint; AData: Byte; AParity: TParity; AStop: Byte);
const
  ParityBits: array[TParity] of Byte = (NO_PARITY, ODD_PARITY, EVEN_PARITY);
var
  Divisor: Word;
  Setting: Byte;
begin
  if (ABps = 0) or
     (AData < 5) or (AData > 8) or
     (AStop < 1) or (AStop > 2) then
    begin
      Error := errBadSettings;
      Exit;
    end;
  Bps := ABps;
  Data := AData;
  Stop := AStop;
  Parity := AParity;
  if Mode = comInt14 then
    begin
      {> FOSSIL? <}
      Setting := Data-5;
      if Stop <> 1 then
        Setting := Setting or $04;
      Setting := Setting or ParityBits[Parity];
      if Bps >= 9600 then Setting := Setting or $E0
      else if Bps >= 4800 then Setting := Setting or $C0
      else if Bps >= 2400 then Setting := Setting or $A0
      else if Bps >= 1200 then Setting := Setting or $80
      else if Bps >= 600 then Setting := Setting or $60
      else if Bps >= 300 then Setting := Setting or $40
      else if Bps >= 150 then Setting := Setting or $20;
      asm
        mov  ah, 0
        mov  al, Setting
        mov  dx, PortNo
        dec  dx
        int  14h
      end;
    end
  else
    begin
      Divisor := 115200 div Bps;
      Setting := Data-5;
      if Stop <> 1 then
        Setting := Setting or $04;
      Setting := Setting or ParityBits[Parity];

      asm cli end;
      if Mode = comPoll then Port[PortBase + IER] := 0 { Disable IRQs }
      else Port[PortBase + IER] := RX_INT;             { Enable IRQs }
      Port[PortBase + MCR] := MC_INT;                  { Enable OUT2 on port }
      Port[PortBase + LCR] := $80;                     { Set DLAB }
      Port[PortBase + DLL] := WordRec(Divisor).Lo;
      Port[PortBase + DLH] := WordRec(Divisor).Hi;
      Port[PortBase + LCR] := Setting;                 { Reset DLAB }

      if Port[PortBase + RXR] = 0 then;  { Remove any pending character }
      if Port[PortBase + LSR] = 0 then;  { Reset line status register }

      Port[PortBase + MCR] := Port[PortBase + MCR] or DTR;     { Enable DTR }
      asm sti end;
    end;
end;

procedure TComPort.InitInt(ABufSize: Word);
begin
  if Mode = comInt then Exit;
  BufSize := ABufSize;
  GetMem(Buffer, BufSize);
  Head := 0; Tail := 0;
  Mode := comInt;
  case IrqNo of
    3:
      begin
        AddToList(Irq3Ports, @Self);
        if OldIsr3 = nil then
          begin
            OldIsr3Enable := IrqEnabled(3);
            GetIntVec($08+3, OldIsr3);
            SetIntVec($08+3, @Isr3);
            EnableIrq(3, True);
          end;
      end;
    4:
      begin
        AddToList(Irq4Ports, @Self);
        if OldIsr4 = nil then
          begin
            OldIsr4Enable := IrqEnabled(4);
            GetIntVec($08+4, OldIsr4);
            SetIntVec($08+4, @Isr4);
            EnableIrq(4, True);
          end;
      end;
  end;
  Port[PortBase + IER] := RX_INT; { Enable IRQs }
end;

procedure TComPort.DoneInt;
begin
  if Mode <> comInt then Exit;
  Port[PortBase + IER] := 0; { Disable IRQs }
  case IrqNo of
    3:
      begin
        RemoveFromList(Irq3Ports, @Self);
        if Irq3Ports = nil then
          begin
            EnableIrq(3, OldIsr3Enable);
            SetIntVec($08+3, OldIsr3);
            OldIsr3 := nil;
          end;
      end;
    4:
      begin
        RemoveFromList(Irq4Ports, @Self);
        if Irq4Ports = nil then
          begin
            EnableIrq(4, OldIsr4Enable);
            SetIntVec($08+4, OldIsr4);
            OldIsr4 := nil;
          end;
      end;
  end;
  Mode := comPoll;
  FreeMem(Buffer, BufSize);
  Buffer := nil;
end;

procedure TComPort.Done;
begin
  DoneInt;
  { Restore port settings }
  asm cli end;
  Port[PortBase + MCR] := OldMCR;
  Port[PortBase + IER] := OldIER;
  Port[PortBase + LCR] := $80;                     { Set DLAB }
  Port[PortBase + DLL] := OldDLL;
  Port[PortBase + DLH] := OldDLH;
  Port[PortBase + LCR] := OldLCR;
  Port[PortBase + MSR] := OldMSR;
  asm sti end;
end;


function TComPort.LineStatus: Byte;
begin
  case Mode of
    comInt14:
      asm
        mov  ax, 0300h
        mov  dx, PortNo
        dec  dx
        mov  @Result, ah
      end;
    else
      LineStatus := Port[PortBase + LSR];
  end;
end;

function TComPort.ModemStatus: Byte;
begin
  case Mode of
    comInt14:
      asm
        mov  ax, 0300h
        mov  dx, PortNo
        dec  dx
        mov  @Result, al
      end;
    else
      ModemStatus := Port[PortBase + MSR];
  end;
end;

function TComPort.Ready: Boolean;
{ Ready to send? }
begin
  Ready := LineStatus and XMTRDY <> 0;
end;

function TComPort.Send(B: Byte): Boolean;
{ Send a byte }
var
  Time: Longint;
begin
  case Mode of
    comInt14:
      asm
        mov  ah, 1
        mov  al, B
        mov  dx, PortNo
        dec  dx
        int  14h
        test ah, 80h
        jz   @Ok
        les  bx, Self
        mov  es:[bx].TComPort.Error, errCantSend
      @Ok:
      end;
    else
      begin
        Time := GetTimer;
        { Wait for transmitter to clear }
        while Port[PortBase + LSR] and XMTRDY = 0 do
          begin
            if (GetTimer >= Time+Timeout) or
               (Assigned(AbortIO) and AbortIO) then { $B+ needed! }
              begin
                Error := errCantSend;
                Send := False;
                Exit;
              end;
            if GetTimer < Time then Time := GetTimer; { Midnight }
          end;

        if HardwareFlow then
          begin
            Port[PortBase + MCR] := MC_INT or DTR or RTS;
            Time := GetTimer;
            { Wait for transmitter to clear }
            while Port[PortBase + MSR] and CTS = 0 do
              begin
                if (GetTimer >= Time+Timeout) or
                   (Assigned(AbortIO) and AbortIO) then { $B+ needed! }
                  begin
                    Error := errCantSend;
                    Send := False;
                    Exit;
                  end;
                if GetTimer < Time then Time := GetTimer; { Midnight }
              end;
          end;

        asm cli end;
        Port[PortBase + TXR] := B;
        asm sti end;
      end;
  end;
  Send := Error = errOk;
end;

function TComPort.Pending: Boolean;
{ Any bytes pending? }
begin
  case Mode of
    comInt:
      Pending := Head <> Tail;
    else
      Pending := LineStatus and RCVRDY <> 0;
  end;
end;

function TComPort.Receive(var B: Byte): Boolean;
{ Receive a byte }
var
  Time: Longint;
begin
  case Mode of
    comInt14:
      asm
        mov  ax, 0200h
        mov  dx, PortNo
        dec  dx
        int  14h
        test ah, 80h
        jz   @Ok
        les  bx, Self
        mov  es:[bx].TComPort.Error, errTimeout
      @Ok:
        les  bx, B
        mov  es:[bx], al
      end;
    comPoll:
      begin
        Time := GetTimer;
        { Wait for transmitter to clear }
        while Port[PortBase + LSR] and RCVRDY = 0 do
          begin
            if (GetTimer >= Time+Timeout) or
               (Assigned(AbortIO) and AbortIO) then { $B+ needed! }
              begin
                Error := errTimeout;
                Receive := False;
                Exit;
              end;
            if GetTimer < Time then Time := GetTimer; { Midnight }
          end;

        asm cli end;
        B := Port[PortBase + RXR];
        asm sti end;
      end;
    comInt:
      begin
        Time := GetTimer;
        { Wait for a character to appear in the buffer }
        while Head = Tail do
          begin
            if (GetTimer >= Time+Timeout) or
               (Assigned(AbortIO) and AbortIO) then { $B+ needed! }
              begin
                Error := errTimeout;
                Receive := False;
                Exit;
              end;
            if GetTimer < Time then Time := GetTimer; { Midnight }
          end;
        B := Buffer^[Head];
        Inc(Head); if Head = BufSize then Head := 0;
      end;
  end;
  Receive := Error = errOk;
end;

var
  OldExit: Pointer;

procedure Finalization; far;
begin
  ExitProc := OldExit;
  if OldIsr3 <> nil then
    begin
      EnableIrq(3, OldIsr3Enable);
      SetIntVec($08+3, OldIsr3);
      OldIsr3 := nil;
    end;
  if OldIsr4 <> nil then
    begin
      EnableIrq(4, OldIsr4Enable);
      SetIntVec($08+4, OldIsr4);
      OldIsr4 := nil;
    end;
end;

begin
  OldExit := ExitProc; ExitProc := @Finalization;
end.
